(use-modules (ice-9 format))


;---------------------------------------
; Reconstruction of Karlheinz Stockhausen's "Studie II" (1954)
; Version 0.1 (2010)
; written by Jan Baumgart (baumgart@folkwang-uni.de)
;
; Released under the GNU General Public License v3 (see gpl.txt)
;
; based on the works of Joachim Heintz and Heinz Silberhorn (Die Reihentechnik in Stockhausens Studie II, Rohrdorfer   ;
;Musikverlag 1980).
;
;
; Currently only Part 1 is implemented
;
;
; Render with CSound:
; csound studie2.orc studie2_teil1.sco -Wdo studie2_teil1.wav
;
;------------------ G L O B A L E  V A R I A B L E N ----------------------------------
(define *pfad* "/Users/jan/Documents/0projekte/ISL/")       ; PFAD fuer Csound Score Output

(define *grund-freq* 100)
;(define *grund-dauer* 1.56139594710949) ;(/ (num->freq 61 2.5 *schritt-faktor*) 76.2)
(define *grund-dauer* 118.978371169743) ; cm

(define *schritt-faktor* (expt 5 1/25)) ; Tonhoehen und Dauern-Temperatur

(define *dynamik* -30) ; in dB

(define *5* 5) ; die magische 5

(define *reihe* '(3 5 1 4 2))    ; die Reihe

(define *-inf* -96) ; stille in db

;------------------ K L E I N E  H E L F E R ----------------------
(define displays
  (lambda args
    (newline)
    (for-each (lambda (x)
		  (display x)
		  (display " "))
	      args)
    (newline)))

(define displays-nl
  (lambda args
    (newline)
    (for-each (lambda (x)
                (display x)
                (newline))
	      args)))

(define (write-sco-line . args)
  (display "i")
  (for-each (lambda (x)
              (display x)
              (display " "))
            args)
  (newline))

(define (flatten ls)
  (cond
   ((null? ls) '())
   ((pair? (car ls))
    (append (flatten (car ls)) (flatten (cdr ls))))
   (else (cons (car ls) (flatten (cdr ls))))))

; modulo, so dass ergebnis zwischen 1..limit(inklusiv) liegt
(define (modlim x limit)
  (+ (modulo (- x 1) limit) 1))

; map von zweiwertiger funktion auf jeweils element und naechste element einer liste
; (list (proc el1 el2) (proc el2 el3) (proc el3 el4))...
(define (map-2 proc ls)
  (map proc
       (list-head ls (- (length ls) 1))
       (cdr ls)))


; /---------------------------------------------\
;|      R E I H E N - G E N E R A T O R E N      |
; \---------------------------------------------/

; transponiert intervall-reihe row auf start als anfangswert
(define (row-transp-to row start limit)
  (let ((delta (- start (car row))))
    (map (lambda (x)
           (modlim (+ x delta) limit))
         row)))

;(row-transp-to '(3 5 1 4 2) 3 5)

; erzeugt ein reihen-quadrat durch transposition von series
; mit series als startwerten (horizontal)
; stocki R1
(define (make-transp-square series limit)
  (define (iter result rest)
    (if (null? rest)
        result
        (iter (append result (list (row-transp-to series (car rest) limit)))
              (cdr rest))))
  (iter '() series))

;(make-number-square (cadr (make-number-square '(3 5 1 4 2) 5)) 5)

; transformiert reihe durch "ueberspringen" von hop-elementen
; hop=1 -> original-reihe
(define (series-hop series hop)
  (define (iter result index count)
    (if (zero? count)
        result
        (let ((newindex (modulo (+ index hop) (length series))))
          (iter (append result (list (list-ref series index)))
                newindex
                (- count 1)))))
  (iter '() 0 (length series)))

; liefert index vom ersten auftreten von el in ls
(define (find-index el ls)
  (define (iter index rest)
    (if (null? rest)
        -1    ; not found
        (if (equal? el (car rest))
            index
            (iter (+ index 1) (cdr rest)))))
  (iter 0 ls))

; liefert schritt-distanz von *from* nach *to*
; innerhalb der reihe *series*
(define (find-step-distance from to series)
  (let ((from-index (find-index from series))
        (to-index (find-index to series)))
    (modulo (- to-index from-index) (length series))))

;(find-step-distance 2 3 '(3 5 1 4 2))

; liefert liste von schritt-distanzen von series
(define (step-distance-ls index-series series)
  (map-2 (lambda (x y)
           (find-step-distance x y index-series))
         series))

; schritt distanzen fuer zahlen-quadrat
(define (step-distance-square square)
  (map (lambda (series)
         (step-distance-ls (car square) series))
       square))

; transformiert *series* durch anwendung der schritte in *step-ls*
; Anfangswert ist *start*
(define (row-trans-step series start step-ls limit)
  (define (iter result index step-rest)
    (if (null? step-rest)
        (append result (list (list-ref series index)))
        (let ((newindex (modulo (+ index (car step-rest)) (length series))))
          (iter (append result (list (list-ref series index)))
                newindex
                (cdr step-rest)))))
    (iter '() (find-index start series) step-ls))

; erzeugt zahlen-quadrat durch anwendung der schritte in step-square auf ori-series
; jeweils erster wert der reihe ist durch start-series festgelegt
(define (make-step-square ori-series start-series step-square limit)
  (map (lambda (startwert step-ls)
         (row-trans-step ori-series startwert step-ls limit))
       start-series step-square))


; Umkehrung von intervall-reihe *series*
(define (invert-series series limit)
  (define (invert x)
    (- (+ limit 1) x))
  (map invert series))

; Umkehrung von reihen-square *square*
(define (invert-square square limit)
  (map (lambda (x)
         (invert-series x limit))
       square))

; Krebs von *series*
(define (reverse-series series)
  (reverse series))

; Krebs von *square*
(define (reverse-square square)
  (map reverse square))

; liefert Wert an position *index* in der Reihe *row*
; aus Quadrat *square*
; row 0..
; square 0..
(define (square-ref row index square)
  (list-ref (list-ref square row) index))


; /-----------------------------------------------------------\
;|      R E I H E N - Z A H L E N  ->  P A R A M E T E R       |
; \-----------------------------------------------------------/


; errechnet die Frequenz vom *num*ten Ton ausgehend
; von *grund-freq* und dem verhaeltnis *schritt-faktor*
(define (num->freq num grund-freq schritt-faktor)
  (* grund-freq (expt schritt-faktor (- num 1))))

(define (gruppe/spalte/stelle->num gruppe spalte stelle limit)
  (+ (* (- gruppe 1) limit) (* (- stelle 1) spalte) 1))


(define (gruppe/spalte/stelle->num-ls gruppe spalte stelle anzahl limit)
  (define (iter num count)
    (if (zero? count)
        '()
        (cons num (iter (+ num spalte) (- count 1)))))
  (let ((erste (gruppe/spalte/stelle->num gruppe spalte stelle limit)))
    (iter erste anzahl)))

(define (gruppe/spalte/stelle->freq-ls gruppe spalte stelle anzahl limit)
  (map (lambda (num)
         (num->freq num *grund-freq* *schritt-faktor*))
       (gruppe/spalte/stelle->num-ls gruppe spalte stelle anzahl limit)))

(define (num->dauer num grund-dauer dauern-faktor)
  (* grund-dauer (expt (/ 1 dauern-faktor) (- num 1))))

(define (gruppe/spalte/stelle->dauer gruppe spalte stelle limit)
  (num->dauer (gruppe/spalte/stelle->num gruppe spalte stelle limit) *grund-dauer* *schritt-faktor*))

; 2do: abstrahieren (wieviel nums fuer best. limit?)
(define (num->db num)
  (let* ((num0 (- num 1))
         (x (if (> num0 30)
                (- 60 num0)
                num0)))
    (* (/ x 30) *dynamik*)))

(define (gruppe/spalte/stelle->db gruppe spalte stelle limit)
  (num->db (gruppe/spalte/stelle->num gruppe spalte stelle limit)))


;----------------------------------------------------------------------------------------

; Erzeugt Ereignisse fuer Teil1 aus den Reihen fuer Gruppen-Transposition, Gruppe, Spalte, Stelle
; Die Ereignise einer Sequenz stehen jeweils in einer Liste
; Ereignis-Parameter:
; (instr time dauer amp env f1 f2 f3 f4 f5)
(define (make-event-ls-teil1 start sequenz
                             freq-trans   freq-gr   freq-sp   freq-st 
                             amp-trans    amp-gr    amp-sp    amp-st
                             dur-trans    dur-gr    dur-sp    dur-st
                             edelay-trans edelay-gr edelay-sp edelay-st
                             env-ls
                             result)
  
  (define (make-seq abstand seq-abstand seq-count
                    f-st a-st d-st ed-st e-ls
                    seq-result)
    (if (zero? seq-count)
        (if (null? (cdr sequenz))                    ; letztes Ereignis in Sequenz ?
            (append result (list seq-result))        ; fertige Sequenz
            (make-event-ls-teil1 (+ start seq-abstand)        (cdr sequenz)
                                 freq-trans   (cdr freq-gr)   (cdr freq-sp)   f-st
                                 amp-trans    (cdr amp-gr)    (cdr amp-sp)    a-st
                                 dur-trans    (cdr dur-gr)    (cdr dur-sp)    d-st
                                 edelay-trans (cdr edelay-gr) (cdr edelay-sp) ed-st
                                 e-ls
                                 (append result (list seq-result))))   ; naechste Stelle
        (let* ((freq-ls (gruppe/spalte/stelle->freq-ls (+ freq-trans   -1 (car freq-gr))   (car freq-sp)   (car f-st)  *5* *5*))
               (amp     (gruppe/spalte/stelle->db      (+ amp-trans    -1 (car amp-gr))    (car amp-sp)    (car a-st)  *5*))
               (dauer   (gruppe/spalte/stelle->dauer   (+ dur-trans    -1 (car dur-gr))    (car dur-sp)    (car d-st)  *5*))
               (ed      (gruppe/spalte/stelle->dauer   (+ edelay-trans -1 (car edelay-gr)) (car edelay-sp) (car ed-st) *5*))
               (env     (car e-ls))
               (time    (+ start abstand))
               (instr 1)
               (event   (append (list instr time dauer amp env) freq-ls)))
          
          (make-seq (+ abstand dauer) (+ seq-abstand ed) (- seq-count 1)  
                    (cdr f-st) (cdr a-st) (cdr d-st) (cdr ed-st) (cdr e-ls)
                    (append seq-result (list event))))))  ; naechste Sequenz
  
  (make-seq 0 0 (car sequenz) freq-st amp-st dur-st edelay-st env-ls '()))


(define (make-event-ls-teil2 start sequenz
                             freq-trans   freq-gr   freq-sp   freq-st 
                             amp-trans    amp-gr    amp-sp    amp-st
                             dur-trans    dur-gr    dur-sp    dur-st
                             edelay-trans edelay-gr edelay-sp edelay-st
                             env-ls
                             result)
  
  (define (make-seq abstand seq-abstand seq-count
                    f-st a-st d-st ed-st e-ls
                    seq-result)
    (if (zero? seq-count)
        (if (null? (cdr sequenz))                    ; letztes Ereignis in Sequenz ?
            (append result (list seq-result))        ; fertige Sequenz
            (make-event-ls-teil1 (+ start seq-abstand)        (cdr sequenz)
                                 freq-trans   (cdr freq-gr)   (cdr freq-sp)   f-st
                                 amp-trans    (cdr amp-gr)    (cdr amp-sp)    a-st
                                 dur-trans    (cdr dur-gr)    (cdr dur-sp)    d-st
                                 edelay-trans (cdr edelay-gr) (cdr edelay-sp) ed-st
                                 e-ls
                                 (append result (list seq-result))))   ; naechste Stelle
        (let* ((freq-ls (gruppe/spalte/stelle->freq-ls (+ freq-trans   -1 (car freq-gr))   (car freq-sp)   (car f-st)  *5* *5*))
               (amp     (gruppe/spalte/stelle->db      (+ amp-trans    -1 (car amp-gr))    (car amp-sp)    (car a-st)  *5*))
               (dauer   (gruppe/spalte/stelle->dauer   (+ dur-trans    -1 (car dur-gr))    (car dur-sp)    (car d-st)  *5*))
               (ed      (gruppe/spalte/stelle->dauer   (+ edelay-trans -1 (car edelay-gr)) (car edelay-sp) (car ed-st) *5*))
               (env     (car e-ls))
               (time    (+ start abstand))
               (instr 1)
               (event   (append (list instr time dauer amp env) freq-ls)))
          
          (make-seq abstand (+ seq-abstand ed) (- seq-count 1)  
                    (cdr f-st) (cdr a-st) (cdr d-st) (cdr ed-st) (cdr e-ls)
                    (append seq-result (list event))))))  ; naechste Sequenz
  
  (make-seq 0 0 (car sequenz) freq-st amp-st dur-st edelay-st env-ls '()))


; env -> amp1 amp2
; verbindet lautstaerke-huellkurve von aufeinanderfolgenden dim. klaenge fuer Teil1
; Ereignis-Parameter:
; Input:  (instr time dauer amp env f1 f2 f3 f4 f5)
; Output: (instr time dauer amp1 amp2 f1 f2 f3 f4 f5)
(define (mold-teil1 event-ls)
  (define (iter event-rest result)
    (if (null? event-rest)
        result
        (iter (cdr event-rest) (append result (seq (car event-rest) '())))))
  (define (seq seq-rest result)
    (if (null? seq-rest)
        result
        (let* ((event   (car seq-rest))
               (instr   (list-ref event 0))
               (time    (list-ref event 1))
               (dauer   (list-ref event 2))
               (amp     (list-ref event 3))
               (env     (list-ref event 4))
               (freq-ls (list-tail event 5)))
          (if (= env 1)     ; cresc
              (seq (cdr seq-rest) (append result (list (append (list instr time dauer *-inf* amp) freq-ls))))   ; cresc
              (if (null? (cdr seq-rest)) ; letztes event in der sequenz ?
                  (seq (cdr seq-rest) (append result (list (append (list instr time dauer amp *-inf*) freq-ls)))) ; diminuendo
                  (let ((amp2 (list-ref (cadr seq-rest) 3))    ; amp von naechstem event
                        (env2 (list-ref (cadr seq-rest) 4)))  ; env von naechstem event
                    (if (= env2 2)  ;wenn naechstes event auch dim. dann verbinden 
                        (seq (cdr seq-rest) (append result (list (append (list instr time dauer amp amp2) freq-ls)))) ; huellkurve verbinden
                        (seq (cdr seq-rest) (append result (list (append (list instr time dauer amp *-inf*) freq-ls))))))))))) ; diminuendo
  (iter event-ls '()))
                    


(define (event-ls->sco event-ls)
  (for-each (lambda (ls)
              (apply write-sco-line ls))
            event-ls))
  

; Quadrat 2-5 erzeugen
(define (make-square-R2-5 r1 num)
  (let* ((steps (step-distance-square r1))
         (row   (list-ref R1 (- num 1))))
    (make-step-square row (series-hop row num) steps *5*)))

; alle Quadrate
(begin
  (define R1 (make-transp-square *reihe* *5*))
  (define R2 (make-square-R2-5 R1 2))
  (define R3 (make-square-R2-5 R1 3))
  (define R4 (make-square-R2-5 R1 4))
  (define R5 (make-square-R2-5 R1 5))
  
  (define U1 (invert-square R1 5))
  (define U2 (invert-square R2 5))
  (define U3 (invert-square R3 5))
  (define U4 (invert-square R4 5))
  (define U5 (invert-square R5 5))

  (define r1 (flatten R1))
  (define r2 (flatten R2))
  (define r3 (flatten R3))
  (define r4 (flatten R4))
  (define r5 (flatten R5))

  (define u1 (flatten U1))
  (define u2 (flatten U1))
  (define u3 (flatten U1))
  (define u4 (flatten U1))
  (define u5 (flatten U1)))

; Huellkurven fuer Teil1 (1=fallend  2=steigend)
(define env1-ls '(
2 1   1 2 2 1   2 2 2 1 2   2 2 1   2
2 1   1 2 2   2   1 2 1 2   2 1 2 1 2
1 2   1 2 2   1 2 2 2   2 1 2 2 2   1
2 2   2   2 1 2 2   1 2 2   1 2 2 1 2
1 2   2   2 1 2 1 2   2 2 1 2   1 2 2
))


(with-output-to-file (string-append *pfad* "studie2_teil1.sco") ; 
  (lambda ()
    (format #t "#include \"studie2.tab\" \n")
   (event-ls->sco
   (mold-teil1
   (make-event-ls-teil1 0 r5
                      3 r1 r2 (append r1 r2 r3) ;freq
                      5 r2 r3 (append r2 r3 r4) ;amp
                      1 r3 r4 (append r3 r4 r5) ;dur
                      1 r4 u5 (append r4 r5 r1) ;ed
                      env1-ls '())))))

;;     (event-ls->sco
;;     (mold-teil1
;;     (make-event-ls-teil2 0 r1
;;                        5 r2 r3 (append r4 r5 r1) ;freq
;;                        2 r3 r4 (append r5 r1 r2) ;amp
;;                        3 r4 r5 (append r1 r2 r3) ;dur
;;                        3 r5 u1 (append r2 r3 r4) ;ed
;;                        env1-ls '())))))




;(num->dauer 6 *grund-dauer* *schritt-faktor*)
;(gruppe/reihe/pos->dauer 2 2 3 5)
;(num->freq 2 2.5 *schritt-faktor*)
;(/ (num->freq 61 2.5 *schritt-faktor*) 76.2)
;(* 100 (expt 5 1/25))
;(reverse-square s1)
;(square-ref 6 2 s1)
;(invert-series '(3 5 1 4 2) 5)
;(row-trans-step '(5 2 3 1 4) 3 (step-distance-ls '(3 5 1 4 2) '(5 2 3 1 4)) 5)
;(row-trans-step '(1 3 4 2 5) 2 (step-distance-ls '(3 5 1 4 2) '(5 2 3 1 4)) 5)
;(row-trans-step '(1 2 4 3 5) 5 '(
;(step-distance-ls '(3 5 1 4 2) '(5 2 3 1 4))
;(step-distance-ls '(1 3 4 2 5) '(2 3 4 5 1))
;(row-trans-step '(1 3 4 2 5) 2 (step-distance-ls '(3 5 1 4 2) '(1 3 4 2 5)) 5)


;(apply displays-nl
; (make-step-square '(5 2 3 1 4) (series-hop '(5 2 3 1 4) 2) (step-distance-square R1) 5))

;(apply displays-nl
; (make-step-square '(2 5 4 3 1) '(2 2 2 2 2) (step-distance-square R1) 5))

;(step-distance-square '((2 5 4 3 1) (2 1 5 4 3) (2 3 1 4 5) (2 5 1 3 4) (2 3 5 1 4)))



; Huellkurven Teil1 nach 
; 2 2 1 2 2
;
; 21 21 12 22 12
;
; 221 122 122 122 122
;
; 1221 1212 1222 2122 2122
;
; 22212 21212 21222 12212 21212

